/********************************************************

	Exception object
	Copyright 2000 Eugene Smith (divx@euro.ru)
	Last modified: 01.08.2000

*********************************************************/
#ifndef __except_h
#define __except_h

#include <iostream.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>

/**
 *
 *   The idea is to use error codes for non-fatal problems
 * ( for example, problem in decompressing one frame is not fatal
 * for video decoder ) and exceptions for problems which leave
 * object in unusable state. GenError is an intermediate case.
 *
 */

using namespace std;

class BaseError
{
    char* module;
    char* description;
    const char* file;
    char* severity;
    int line;
public:
    ~BaseError()
    {
	delete module;
	delete description;
	delete severity;
    }	
    BaseError(const char* sev, const char* mod, const char* desc, const char* f=0, int l=0):file(f),line(l)
    {
	module=new char[strlen(mod)+1];
	strcpy(module, mod);
	severity=new char[strlen(sev)+1];
	strcpy(severity, sev);
	description=new char[strlen(desc)+1];
	strcpy(description, desc);
    }
    BaseError(const BaseError& f)
    {
	module=new char[strlen(f.module)+1];
	strcpy(module, f.module);
	description=new char[strlen(f.description)+1];
	strcpy(description, f.description);
	severity=new char[strlen(f.severity)+1];
	strcpy(severity, f.severity);
	file=f.file;
	line=f.line;
    }
    void Print()
    {
	cerr<<module<<": ";
	cerr<<severity;
	cerr<<": "<<description<<endl;
    }
    void PrintAll()
    {
	cerr<<module<<": ";
	cerr<<severity;
	cerr<<": "<<description;
	if(file)cerr<<" at "<<file;
	if(line)cerr<<":"<<line;
	cerr<<endl;
    }		
    const char* GetModule() const{return module;}
    const char* GetDesc() const{return description;}
};
class FatalError: public BaseError
{
public:
    FatalError(const char* mod, const char* desc, const char* f=0, int l=0):BaseError("FATAL",mod,desc,f,l){}
    FatalError(const FatalError& f):BaseError(f){}
};
class GenError: public BaseError
{
public:
    GenError(const char* mod, const char* desc, const char* f=0, int l=0):BaseError("WARNING",mod,desc,f,l){}
    GenError(const GenError& f):BaseError(f){}
};

#define FATAL(X) FatalError(__MODULE__,X,__FILE__,__LINE__)
#define WARNING(X) GenError(__ERR_MODULE__,X,__FILE__,__LINE__)
#define MEMORY_FATAL FATAL("Out of memory")

#endif