/********************************************************

	Video encoder interface
	Copyright 2000 Eugene Smith (divx@euro.ru)
	Last modified: 28.07.2000

*********************************************************/


#include <stdio.h>
#include <math.h>

#include <loader.h>

#include <default.h>
#include <except.h>
/**
 *
 *    Video encoder class.
 *
 *  Usage ( preliminary info ):
 *	
 *   Create encoder object. Call Init(), optionally
 * set quality and key-frame frequency. Call Start().
 * Call QueryOutputSize() to identify maximum possible
 * size of compressed frame. Allocate buffer of this size.
 * Pass it as 'dest' in calls to EncodeFrame(). At the end
 * call Stop().
 *
 *  Some codec DLLs have be controlled by special codec-specific
 * parameters. In Windows you can set them from codec configuration 
 * dialog, but since we are completely GUI-independent, we'll need
 * an internal method to set them. Currently it is done by calling
 * static func SetExtendedAttr(), which takes FOURCC of codec, 
 * name of parameter and its integer value. Complete list of
 * supported attributes can be found in source of function
 * SetExtendedAttr() ;-) This func should be called before Init()
 * because values for these attributes are typically stored in 
 * registry and read by codec during object creation.
 *
 *  The complete scheme looks extremely ugly, but it is in very alpha.
 * This class was actually written in half an hour at the very beginning
 * of Avifile and almost didn't change since then. Some day I'll
 * return here and fix everything.
 *
 */
class VideoEncoder
{
    HIC hic;
 // the heart of encoder
    int m_comp_id;    
    int m_iState;
    BITMAPINFOHEADER m_bh, m_obh;
    int m_iFrameNum;
    int m_iQual;
    int m_iKfFreq;
    char* m_prev;
    //following are DivX-specific values
    int m_dBitRate;
public:
    VideoEncoder():m_iState(0), m_dBitRate(0), m_prev(0){}
    ~VideoEncoder();
    //
    //  Should be called before operation
    //  Specifies source video format and compressor
    //
    void Init(int compressor, const char*) throw(FatalError&);//BITMAPINFOHEADER*
    void Close();
    //
    // Starts encoding
    //
    void Start();
    //
    // Stops encoding
    //
    void Stop();
    //
    // Queries encoder about desired buffer size for compression
    // You should allocate at least this much bytes for dest ( arg for EncodeFrame )
    //
    int QueryOutputSize();
    BITMAPINFOHEADER QueryOutputFormat() const
    {
	return m_obh;
    }	
    //
    // Encodes single frame
    // On success, is_keyframe contains 0 if frame was encoded into delta frame, otherwise AVIIF_KEYFRAME (?)
    // size receives size of compressed frame
    //
    int EncodeFrame(char* src, char* dest, int* is_keyframe, int* size, int* lpckid=0);
    //
    // Quality takes values from 0 to 10000, 10000 the best
    //
    int SetQuality(int quality);
    int GetQuality();
    int SetKeyFrame(int frequency);
    int GetKeyFrame();
    static int SetExtendedAttr(int id, char* attribute, int value);
protected:  
    static int GetRegValue(int fccHandler, const char* name, int* place);                      
    static int SetRegValue(int fccHandler, const char* name, int value);                      
};    
    