/********************************************************

	Classes for reading AVIs
	Copyright 2000 Eugene Smith (divx@euro.ru)
	Last modified: 30.07.2000

*********************************************************/


#include <avifmt.h>
#include <default.h>
#include <unistd.h>
#include <malloc.h>
#include <string.h>
#include <stdio.h>
#include <pthread.h>
#include "AVIReadHandler.h"
#include "AVIReadHandler2.h"
#include <avifile.h>
#include <audiodecoder.h>
#include <videodecoder.h>    
#include <vector>
using namespace std;

class AviReadStream : public IAviReadStream
{
friend class AviReadFile;

protected:
    IvAVIReadStream* m_pIStream;
    AVIIndexEntry2* m_pTable;
    AVIStreamHeader m_header;
    unsigned m_id;
    char* m_format;
    long m_format_size;
    unsigned m_lockcount;
    
    unsigned m_position;
    unsigned m_ch_off;//chunk offset for streams with fixed sample size
    
    unsigned m_sample;//for streams with fixed sample size: total number of samples sent
    
    unsigned m_table_size;
    offset_t start_offset;
    int m_streaming;


    char rem_buffer[10000];
    int rem_size;    
    BITMAPINFOHEADER m_bh;
    BITMAPINFOHEADER m_obh;
public:
//AviReadStream-specific
    AviReadStream():rem_size(0), m_pTable(0), m_pIStream(0), m_sample(0), m_ch_off(0), m_position(0),
	m_format_size(0), m_streaming(0){}
    HRESULT Init(unsigned id, IvAVIReadStream* stream);
//Generic
    ~AviReadStream();
    StreamType GetType() const;
    HRESULT GetVideoFormatInfo(void* bi);
    HRESULT GetAudioFormatInfo(void* bi, char** ext);
//Size of stream and one frame
    unsigned GetEndPos() const {return m_header.dwLength;}
    double GetEndTime() const;
    double GetFrameTime() const {return (double)m_header.dwScale / m_header.dwRate;}
//Positioning in stream
    HRESULT Seek(unsigned pos);
    double GetTime() const;
    unsigned GetPos() const;
    HRESULT GetFrameFlags(int* flags);

//Reading decompressed data    
    int IsStreaming();

    HRESULT SetBitDepth(unsigned depth);
    HRESULT SetOutputFormat();
    HRESULT ReadFrame();	
    char* GetFrame();		
    int GetFrameSize();
    
    HRESULT ReadDirect(char* buffer, unsigned bufsize, unsigned samples,
	unsigned& samples_read, unsigned& bytes_read);

    unsigned GetNextKeyFrame() const;
    unsigned GetPrevKeyFrame() const;

    int ToNextKeyFrame();
    int ToPrevKeyFrame();

    HRESULT Eof() const;
};

class AviReadStreamV : public AviReadStream
{
    VideoDecoder* videodecoder;
public:
    AviReadStreamV():AviReadStream(){}
    double SeekToTime(double time);
    HRESULT GetVideoFormatInfo(void* bi);
    HRESULT StartStreaming();
    HRESULT StopStreaming();
    HRESULT SetBitDepth(unsigned depth);
    HRESULT GetOutputFormat(void* format, unsigned size);
    HRESULT ReadFrame();
    int GetFrameSize();
    HRESULT SetOutputFormat();
    char* GetFrame();		
    HRESULT ReadFrames(char* buffer, unsigned bufsize, unsigned samples,
	unsigned& samples_read, unsigned& bytes_read);

};

class AviReadStreamA : public AviReadStream
{
    AudioDecoder* audiodecoder;
public:
    AviReadStreamA():AviReadStream(){}
    double SeekToTime(double time);
    HRESULT GetAudioFormatInfo(void* bi, char** ext);
    unsigned GetPos() const;
    HRESULT StartStreaming();
    HRESULT StopStreaming();
    HRESULT GetOutputFormat(void* format, unsigned size);
    HRESULT ReadFrames(char* buffer, unsigned bufsize, unsigned samples,
	unsigned& samples_read, unsigned& bytes_read);
};

    

class AviReadFile: public IAviReadFile
{
private:
    MainAVIHeader m_header;
    IvAVIReadHandler* m_pIFile;
//    AviReadStream* m_streams;
    vector<AviReadStream*> m_streams;
    int m_vcount;
    int m_acount;
    void _destruct();
public:
//    AviReadFile():m_pIFile(0), m_streams(0){}
    AviReadFile(const char* name);
    ~AviReadFile();
//    HRESULT OpenFile(const char* name);
    unsigned StreamCount();
    unsigned VideoStreamCount();
    unsigned AudioStreamCount();
    IAviReadStream* GetStream(unsigned stream_id, AviStream::StreamType type);
    HRESULT GetFileHeader(MainAVIHeader* header);
    
//    HRESULT CloseFile();
};

