#ifndef _AUDIO_H
#define _AUDIO_H

#include <avifile.h>
#include <except.h>
using namespace std;
class audio_queue
{
public:
    audio_queue();
    ~audio_queue();
    int push(const char* data, int size);
    int write(int fd);
    void clear();
    int size() const {return frame_size;}
    int wait() //wait until there are any data in buffer
    {
	pthread_mutex_lock(&mutex);
	if(frame_size!=0)
	{
	    pthread_mutex_unlock(&mutex);
	    return 0;
	}    
	pthread_cond_wait(&cond, &mutex);
	pthread_mutex_unlock(&mutex);
	return -1;
    }
    void broadcast() {pthread_cond_broadcast(&cond);}
protected:
    pthread_mutex_t mutex;	// Controls access to audio buffer
    pthread_cond_t cond;	// Broadcast to notify audio_thread about
				// new available data OR about quit=1
     
    int frame_size;		// Current amount of data
    static const int frame_max=400000; 
				// Max allowed frame_size
    volatile int frame_in;	
    volatile int frame_out;	// Input and output positions in queue
    unsigned char audio_frame[2*frame_max];
				// Audio queue

};
class AudioRenderer
{
    friend void* doAudioOut(void*);
public:
    double getTime();
    double getLength(){if(!audiostream)return 0; return audiostream->GetEndTime();}
    void doAudioExtract(double);
    AudioRenderer(IAviReadStream* stream);
    ~AudioRenderer();
    void reseek(double pos);
    void start();
    void stop();
    void pause(bool state)
    {
	if(!initialized)return;
        if(!state)time_start=0;
        paused=state;
    }
    double buffer_time();	//for how long can we play without doAudioExtract()'s?
    void clear();		//remove all data from buffers
    void wake();		//restart thread if it waits
    int Eof(){if(audiostream==0)return 0;return audiostream->Eof();}
protected:
    pthread_mutex_t mutex;
    IAviReadStream* audiostream;
    audio_queue queue;
    int paused;			// 1 if we are paused
    int hangup;
    int initialized;		// 1 if we are playing OR paused
    
    WAVEFORMATEX audio;		// format of audio stream
    char audio_ext[100];
    
    int audio_fd;		// /dev/dsp or ESD socket ( syntactically the same behaviour )
    int esd;

    unsigned Snd_Limit;		// DMA buffer size
    WAVEFORMATEX owf;		// Format of data we write to audio_fd
    
    long long time_start;	// Timestamp of moment when we started playing
				// updated after each reseek or pause
    long long audio_time;	// Timestamp of moment when last write to audio_fd
				// has finished
    double audio_realpos;	// Precise position of audio at audio_time

    int debug_status; 		// used for debugging console output
	 
    int quit;			// nonzero value signals to thread
				// to terminate
				
    pthread_t audio_thread;	// Writes data to audio_fd
    
};
#endif