/********************************************************

	Video decoder implementation
	Copyright 2000 Eugene Smith (divx@euro.ru)
	Last modified: 04.08.2000

*********************************************************/


#include <videodecoder.h>
#include <string.h>
#include <except.h>
#define __MODULE__ "VideoDecoder"

#include <errno.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <malloc.h>
#include <stdio.h>
#include <unistd.h>
#ifdef __linux__
#include <asm/unistd.h>
#include <asm/ldt.h>
#else
#define LDT_ENTRIES     8192
#define LDT_ENTRY_SIZE  8
#pragma pack(4)
struct modify_ldt_ldt_s {
        unsigned int  entry_number;
        unsigned long base_addr;
        unsigned int  limit;
        unsigned int  seg_32bit:1;
        unsigned int  contents:2;
        unsigned int  read_exec_only:1;
        unsigned int  limit_in_pages:1;
        unsigned int  seg_not_present:1;
        unsigned int  useable:1;
};

#define MODIFY_LDT_CONTENTS_DATA        0
#define MODIFY_LDT_CONTENTS_STACK       1
#define MODIFY_LDT_CONTENTS_CODE        2
#define __NR_modify_ldt         123
#endif


/**
 *
 *  This should be performed before we create first thread. See remarks
 *  for write_ldt(), linux/kernel/ldt.c.
 *
 */

class LDT_Keeper
{
    void* fs_seg;
public:
    LDT_Keeper();
    ~LDT_Keeper();
    static int modify_ldt( int func, struct modify_ldt_ldt_s *ptr,
                                  unsigned long count )
    {
	int res;
#ifdef __PIC__
	__asm__ __volatile__( "pushl %%ebx\n\t"
                          "movl %2,%%ebx\n\t"
                          "int $0x80\n\t"
                          "popl %%ebx"
                          : "=a" (res)
                          : "0" (__NR_modify_ldt),
                            "r" (func),
                            "c" (ptr),
			    "d"(16)
			  :"esi"     );
#else
        __asm__ __volatile__("int $0x80"
                         : "=a" (res)
                         : "0" (__NR_modify_ldt),
                           "b" (func),
                           "c" (ptr),
			    "d"(16)
			   :"esi");
#endif  /* __PIC__ */
	if (res >= 0) return res;
	errno = -res;
        return -1;
    }
    static void LDT_EntryToBytes( unsigned long *buffer, const struct modify_ldt_ldt_s *content )
    {
	*buffer++ = ((content->base_addr & 0x0000ffff) << 16) |
             (content->limit & 0x0ffff);
        *buffer = (content->base_addr & 0xff000000) |
              ((content->base_addr & 0x00ff0000)>>16) |
              (content->limit & 0xf0000) |
              (content->contents << 10) |
              ((content->read_exec_only == 0) << 9) |
              ((content->seg_32bit != 0) << 22) |
              ((content->limit_in_pages != 0) << 23) |
              0xf000;
    }
};

LDT_Keeper::LDT_Keeper()
{
    struct modify_ldt_ldt_s array;
    int fd;
    int ret;
    void* prev_struct;
    
    fd=open("/dev/zero", O_RDWR);
    fs_seg=mmap((void*)0xbf000000, 0x30000, PROT_READ | PROT_WRITE, MAP_PRIVATE,
	fd, 0);
    if(fs_seg==0)
    {
	printf("ERROR: Couldn't allocate memory for fs segment\n");
	return;
    }	
    array.base_addr=((int)fs_seg+0xffff) & 0xffff0000;
    array.entry_number=0x1;
    array.limit=array.base_addr+getpagesize()-1;
    array.seg_32bit=1;
    array.read_exec_only=0;
    array.seg_not_present=0;
    array.contents=MODIFY_LDT_CONTENTS_DATA;
    array.limit_in_pages=0;
#ifdef linux
    ret=modify_ldt(0x1, &array, 1);
    if(ret<0)
    {
	perror("install_fs");
	printf("Couldn't install fs segment, expect segfault\n");
    }	
#endif /*linux*/

#if defined(__NetBSD__) || defined(__FreeBSD__) || defined(__OpenBSD__)
    {
        long d[2];

        LDT_EntryToBytes( d, &array );
        ret = i386_set_ldt(0x1, (union descriptor *)d, 1);
        if (ret < 0)
        {
            perror("install_fs");
            MESSAGE("Did you reconfigure the kernel with \"options USER_LDT\"?\n");
        }
    }
#endif  /* __NetBSD__ || __FreeBSD__ || __OpenBSD__ */
    __asm__ __volatile__
    (
    "movl $0xf,%eax\n\t"
    "movw %ax, %fs\n\t"
    );
    prev_struct=malloc(8);
    *(void**)array.base_addr=prev_struct;
    close(fd);
}
LDT_Keeper::~LDT_Keeper()
{
    if(fs_seg==0)
	return;
    munmap(fs_seg, 0x30000);
}
static LDT_Keeper Keeper;
    
void VideoDecoder::Init(const char* format)
 throw (FatalError&)
{
    try
    {
    if(m_iState>0)
	return;//already initialized
    if(format==0)
	throw FATAL("Bad format");
    memset(&m_cv, 0, sizeof(m_cv));
    memcpy(&m_bh, format, sizeof(m_bh));
    memset(&m_obh, 0, sizeof(m_obh));
    m_obh.biSize=sizeof(m_obh);
    m_cv.cbSize=sizeof(m_cv);
#warning NOT SURE
//
// MPEG-4 format has changed during development. Package
// binaries.zip includes mpg4c32.dll - old MPEG-4 codec
// with unlocked compression - and divxc32.dll,
// which differs from current MPEG-4 only by FOURCC's.
// The problem is to identify which is needed to us,
// because 'old' codec doesn't work with 'new' clips and
// vice versa.
//

// I've got old video with fccHandler mp43 and biCompression MPG4.


    switch(m_bh.biCompression)
    {
	case mmioFOURCC('M', 'P', '4', '1'):
	case mmioFOURCC('M', 'P', '4', '2'):
	case mmioFOURCC('M', 'P', '4', '3'):
	    m_bh.biCompression=mmioFOURCC('d', 'i', 'v', '3');
	    break;
    }	    	
/*	printf("  biSize %d\n", m_bh.biSize);
	printf("  biWidth %d\n", m_bh.biWidth);
	printf("  biHeight %d\n", m_bh.biHeight);
	printf("  biPlanes %d\n", m_bh.biPlanes);
	printf("  biBitCount %d\n", m_bh.biBitCount);
	printf("  biCompression %d='%.4s'\n", m_bh.biCompression, &m_bh.biCompression);
	printf("  biSizeImage %d\n", m_bh.biSizeImage);*/

    m_cv.hic=ICOpen(
    	0x63646976,//vidc
	m_bh.biCompression,
	ICMODE_FASTDECOMPRESS);
    if(m_cv.hic==0)
	throw FATAL("ICOpen failed");
    
    HRESULT h=ICDecompressGetFormat(m_cv.hic, &m_bh, &m_obh);
    if(h!=0)
    {
	printf("Declined format dump:\n");
	printf("  biSize %d\n", m_bh.biSize);
	printf("  biWidth %d\n", m_bh.biWidth);
	printf("  biHeight %d\n", m_bh.biHeight);
	printf("  biPlanes %d\n", m_bh.biPlanes);
	printf("  biBitCount %d\n", m_bh.biBitCount);
	printf("  biCompression %d='%.4s'\n", m_bh.biCompression, &m_bh.biCompression);
	printf("  biSizeImage %d\n", m_bh.biSizeImage);
	throw FATAL("ICDecompressGetFormat failed");
    }
    m_outFrame=new char[m_obh.biSizeImage?m_obh.biSizeImage:m_obh.biWidth*m_obh.biHeight*m_obh.biBitCount/8];
    if(m_outFrame==0)
	throw FATAL("Memory failure");
    m_iState=1;		
    return;
    }
    catch(FatalError& error)
    {
	if(m_cv.hic)
	    ICClose(m_cv.hic);
	if(m_outFrame)
	    delete m_outFrame;
	m_outFrame=0;
	m_cv.hic=0;
	throw;	
    }	    
}
void VideoDecoder::Start()
{
    if(m_iState!=1)
	return;
/*  printf("Starting decompression, format: \n");
	printf("  biSize %d\n", m_bh.biSize);
	printf("  biWidth %d\n", m_bh.biWidth);
	printf("  biHeight %d\n", m_bh.biHeight);
	printf("  biPlanes %d\n", m_bh.biPlanes);
	printf("  biBitCount %d\n", m_bh.biBitCount);
	printf("  biCompression %d='%.4s'\n", m_bh.biCompression, &m_bh.biCompression);
	printf("  biSizeImage %d\n", m_bh.biSizeImage);
    printf("Dest fmt:\n");
	printf("  biSize %d\n", m_obh.biSize);
	printf("  biWidth %d\n", m_obh.biWidth);
	printf("  biHeight %d\n", m_obh.biHeight);
	printf("  biPlanes %d\n", m_obh.biPlanes);
	printf("  biBitCount %d\n", m_obh.biBitCount);
	printf("  biCompression %d='%.4s'\n", m_obh.biCompression, &m_obh.biCompression);
	printf("  biSizeImage %d\n", m_obh.biSizeImage);*/
    HRESULT hr=ICDecompressBegin(m_cv.hic, &m_bh, &m_obh);
    if(hr!=0)
    {
	cerr<<"WARNING: ICDecompressBegin() failed ( shouldn't happen ), error code "<<hr<<endl;
    }
    m_iState=2;
    return;		
}
void VideoDecoder::Stop()
{
    if(m_iState!=2)
	return;
    HRESULT hr=ICDecompressEnd(m_cv.hic);
    if(hr!=0)
    {
	cerr<<"WARNING: ICDecompressEnd() failed ( shouldn't happen ), error code "<<hr<<endl;
    }
    m_iState=1;
    return;
}
void VideoDecoder::Close()
{
    if(m_iState==0)
	return;
    if(m_cv.hic)
	ICClose(m_cv.hic);
    if(m_outFrame)
	delete m_outFrame;
    m_outFrame=0;
    m_iState=0;
}
int VideoDecoder::DecodeFrame(char* src, int size, int is_keyframe)
{
	m_bh.biSizeImage=size;
	int hr=ICDecompress(m_cv.hic, is_keyframe?0:ICDECOMPRESS_NOTKEYFRAME,
		 &m_bh, src, &m_obh, m_outFrame);
	if(hr!=0)cerr<<"VideoDecoder: warning: hr="<<hr<<endl;
	return hr;
}
int VideoDecoder::SetBitDepth(int bits)
{
    if(m_iState!=1)
	return -1;
    BITMAPINFOHEADER temp=m_obh;
    temp.biBitCount=bits;
    temp.biSizeImage=m_obh.biWidth*m_obh.biHeight*temp.biBitCount/8;    
    HRESULT h=ICDecompressQuery(m_cv.hic, &m_bh, &temp);
    if(h!=0)
	return -1;
    
    char* outFrame=new char[m_obh.biWidth*m_obh.biHeight*bits/8];
    if(m_outFrame)delete m_outFrame;
    m_outFrame=outFrame;	
    m_obh=temp;
    m_bh.biBitCount=bits;
    return 0; 
}
