/********************************************************

	555->565 weight conversion routine
	Copyright 2000 Eugene Smith (divx@euro.ru)
	Last modified: 29.07.2000

*********************************************************/

#ifndef MMX
extern void v555to565(short* dest, const short* src, int w, int h)
{
    dest+=w*(h-1);
    for(int i=0; i<h; i++)
    {
    	for(int j=0; j<w; j++)
	{
	    short q=*src++;
	    q+=(q&0xFFE0);
	    *dest++=q;
	}
	dest-=2*w;
    }
}
extern void zoom_2(short* dest, const short* src, int w, int h)
{
    for(int i=0; i<h; i+=2)
    {
    	for(int j=0; j<w; j+=2)
    {
		*dest=*src;
		dest++;
		src+=2;
	}
	src+=w;
	}
}
extern void zoom_2_to565(short *dest, const short *src, int w, int h)
{
    dest+=w/2*(h/2-1);
    for(int i=0; i<h/2; i++)
    {
	for(int j=0; j<w/2; j++)
        {
    	    short q=*src;
	    q+=(q&0xFFE0);
	    *dest++=q;
            src+=2;
        }
	src+=w;
	dest-=w;
    }
}

#else
/*********************************************

	        WARNING

All MMX code assumes that dest scanline sizes
	are multiple of 8 bytes.		

*********************************************/
extern void v555to565(short* dest, const short* src, int w, int h)
{
    long long line=0xFFe0FFe0FFe0FFe0LL;	
    char qw[200];
    __asm__ __volatile__ 
    (
    "fsave (%0)\n\t"
    "emms\n\t"
    :
    :"r"(&qw));
    dest+=w*(h-1);
    for(int i=0; i<h; i++)
    {
    __asm__ __volatile__
    (
    "pushl 	%%ecx		\n\t"
    "pushl 	%%edx		\n\t"
    
    "movq       %3,     %%mm2   \n\t"
    "movl       %2,     %%edx   \n\t"
    "addl       %%eax,  %%edx   \n\t"
//loop:
    "pos0:\n\t"

    "movq       (%%eax),%%mm0   \n\t"
    "movq       %%mm0,  %%mm1   \n\t"
    "pand       %%mm2,  %%mm1   \n\t"
    "paddw      %%mm1,  %%mm0   \n\t"
    "movq       %%mm0,  (%%ecx) \n\t"
    "addl       $8,     %%eax   \n\t"
    "addl       $8,     %%ecx   \n\t"
    "cmpl       %%eax,  %%edx   \n\t"
    "ja 	pos0		\n\t"

    "popl 	%%edx		\n\t"
    "popl 	%%ecx		\n\t"
    :
    :"a"(src), "c"(dest), "m"(2*w), "m"(line)
    );
    src+=w;
    dest-=w;
    }
    __asm__ __volatile__ ("frstor (%0)\n\t": :"r"(&qw));
}
extern void zoom_2_to565(short *dest, const short *src, int w, int h)
{
    long long line=0xFFe0FFe0FFe0FFe0LL;	
    long long line2=0x0000FFFF0000FFFFLL;	
    char qw[200];
    __asm__ __volatile__ 
    (
    "fsave (%0)\n\t"
    "emms\n\t"
    : 
    :"r"(&qw)
    );//who knows what this crazy compiler decides to store in FPU stack? 

    dest+=w/2*(h/2-1);
    for(int i=0; i<h/2; i++)
    {
	    __asm__ __volatile__
	    (
	        "pushl 	%%ecx		\n\t"
	        "pushl 	%%edx		\n\t"
    
	    	"movq 	%2, 	%%mm2	\n\t"
		"movq 	%3, 	%%mm3	\n\t"
		"movl   %4,     %%edx   \n\t"
    		"addl   %%eax,  %%edx   \n\t"

	        "pos1:			\n\t"
//Load 8 subsequent pixels into mm0 and mm1. 
//Drop each second one by pand.
	        "movq   (%%eax),%%mm0   \n\t"
		"addl 	$8,     %%eax   \n\t"
		"pand 	%%mm3, 	%%mm0	\n\t"
	        "movq   (%%eax),%%mm1   \n\t"
		"addl   $8,     %%eax	\n\t"
		"pand 	%%mm3, 	%%mm1	\n\t"
//Pack 4 remaining pixels into mm0.
		"packssdw %%mm1,%%mm0   \n\t"
//Convert 555 -> 565.	        
		"movq 	%%mm0,  %%mm1   \n\t"
	        "pand   %%mm2,  %%mm1   \n\t"
		"paddw  %%mm1, 	%%mm0 	\n\t"
//Store the result.		
	        "movq   %%mm0, (%%ecx) 	\n\t"

		"addl 	$8,	%%ecx	\n\t"
	        "cmpl   %%eax,  %%edx   \n\t"
	        "ja 	pos1		\n\t"

	        "popl 	%%edx		\n\t"
	        "popl 	%%ecx		\n\t"
	    :
	    :"a"(src), "c"(dest), "m"(line), "m"(line2), "m"(2*w)
	    );
	src+=2*w;
	dest-=w/2;
    }
    __asm__ __volatile__ ("frstor (%0)\n\t": :"r"(&qw));
}
extern void zoom_2_to555(short *dest, const short *src, int w, int h)
{
    long long line2=0x0000FFFF0000FFFFLL;
    char qw[200];
    __asm__ __volatile__
    (
    "fsave (%0)\n\t"
    "emms\n\t"
    :
    :"r"(&qw)
    );

    for(int i=0; i<h/2; i++)
    {
	    __asm__ __volatile__
	    (
	        "pushl 	%%ecx		\n\t"
	        "pushl 	%%edx		\n\t"
    
		"movq 	%2, 	%%mm3	\n\t"
		"movl   %3,     %%edx   \n\t"
    		"addl   %%eax,  %%edx   \n\t"

	        "pos2:			\n\t"
//Load 8 subsequent pixels into mm0 and mm1. 
//Drop each second one by pand.
	        "movq   (%%eax),%%mm0   \n\t"
		"addl 	$8,     %%eax   \n\t"
		"pand 	%%mm3, 	%%mm0	\n\t"
	        "movq   (%%eax),%%mm1   \n\t"
		"addl   $8,     %%eax	\n\t"
		"pand 	%%mm3, 	%%mm1	\n\t"
//Pack 4 remaining pixels into mm0.
		"packssdw %%mm1,%%mm0   \n\t"
//Store the result.		
	        "movq   %%mm0, (%%ecx) 	\n\t"

		"addl 	$8,	%%ecx	\n\t"
	        "cmpl   %%eax,  %%edx   \n\t"
	        "ja 	pos2		\n\t"

	        "popl 	%%edx		\n\t"
	        "popl 	%%ecx		\n\t"
	    :
	    :"a"(src), "c"(dest),  "m"(line2), "m"(2*w)
	    );
	src+=2*w;
	dest+=w/2;
    }
    __asm__ __volatile__ ("frstor (%0)\n\t": :"r"(&qw));
}

extern void zoom_2_32(int *dest, const int *src, int w, int h)
{
    long long line2=0x00000000FFFFFFFFLL;	
    char qw[200];
    __asm__ __volatile__ 
    (
    "fsave (%0)\n\t"
    "emms\n\t"
    : 
    :"r"(&qw)
    );

//    dest+=w/2*(h/2-1);
    for(int i=0; i<h/2; i++)
    {
	    __asm__ __volatile__
	    (
	        "pushl 	%%ecx		\n\t"
	        "pushl 	%%edx		\n\t"
    
		"movl   %2,     %%edx   \n\t"
    		"addl   %%eax,  %%edx   \n\t"

	        "pos3:			\n\t"
//Load 4 subsequent pixels into mm0 and mm1. 
	        "movq   (%%eax),%%mm0   \n\t"
		"addl 	$8,     %%eax   \n\t"
	        "movq   (%%eax),%%mm1   \n\t"
		"addl   $8,     %%eax	\n\t"
//Put 2 pixels into mm0.
//this should work, but I'm too lazy to check it.
		"punpckhdq %%mm1,%%mm0   \n\t"
//Store the result.		
	        "movq   %%mm0, (%%ecx) 	\n\t"

		"addl 	$8,	%%ecx	\n\t"
	        "cmpl   %%eax,  %%edx   \n\t"
	        "ja 	pos3		\n\t"

	        "popl 	%%edx		\n\t"
	        "popl 	%%ecx		\n\t"
	    :
	    :"a"(src), "c"(dest), "m"(4*w)
	    );
	src+=2*w;
	dest+=w/2;
    }
    __asm__ __volatile__ ("frstor (%0)\n\t": :"r"(&qw));
}

#endif
//sloooow....
extern void zoom_2(short* dest, const short* src, int w, int h, int size)
{        
	int* s, *d;           
	switch(size)
	{
	case 15:
	case 16:
#ifdef MMX
	    zoom_2_to555(dest, src, w, h);
#else	    	
		for(int i=0; i<h/2; i++)
    		{
    		for(int j=0; j<w/2; j++)
    		{
			*dest=*src;
			dest++;
			src+=2;
		}
		src+=w;
		}
#endif
		break;
	case 32:
#ifdef MMX
		zoom_2_32((int*)dest, (const int*)src, w, h);	
#else	    
		s=(int*)src;
		d=(int*)dest;
		for(int i=0; i<h/2; i++)
    		{
    		for(int j=0; j<w/2; j++)
    		{
			*d=*s;
			d++;
			s+=2;
		}
		s+=w;
	   	}
#endif		
		break;
	}
}