/********************************************************

	XMITSHM video renderer
	Copyright 2000 Eugene Smith (divx@euro.ru)
	Last modified: 27.07.2000

*********************************************************/
#include <config.h>

#include <qwidget.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/extensions/XShm.h>

#ifdef HAVE_LIBXXF86DGA
#include <X11/extensions/xf86dga.h>
#endif

#include <sys/ipc.h>
#include <sys/shm.h>                               

#include <malloc.h>
#include <pthread.h>
#include "stretch.h"
#include "renderer.h"
#include <except.h>

#define __MODULE__ "VideoRenderer"
int width, height;
//code taken from kwintv
int GetPhysicalDepth(QPaintDevice* dev)
{
        Display *dpy = dev->x11Display();
	int n, pixmap_bits;
	XPixmapFormatValues *pf=XListPixmapFormats(dpy, &n);
	for ( int i = 0; i<n; i++)
	{
		if(pf[i].depth==QColor::numBitPlanes())
		pixmap_bits=pf[i].bits_per_pixel;
	}
	return pixmap_bits;
}

inline static void qSafeXDestroyImage( XImage *x )
{
    if ( x->data ) {
	free( x->data );
	x->data = 0;
    }
    XDestroyImage( x );
}



class ShmRenderer : public VideoRenderer
{
    bool	       xshminit;
    XShmSegmentInfo xshminfo;
    XImage	      *xshmimg;
    Pixmap	       xshmpm;
    GC 			gc;
    int zoom_divider;
    int xpos;
//    int phx, phy;   
//    pthread_mutex_t mutex; 
public:
    ShmRenderer() : xshminit(0), xshmimg(0), xshmpm(0){}
    int Create(QWidget* w, int x, int y, int zoom=1, int xpos=0)
    {
        static int major, minor;
	static Bool pixmaps_ok;
        dev=(QPaintDevice*)w;
        Display *dpy = dev->x11Display();
        int dd	 = dev->x11Depth();
        Visual *vis	 = (Visual*)dev->x11Visual();
	zoom_divider=zoom;
	if(zoom!=1 && zoom!=2)throw FATAL("Unsupported zoom factor");
	m_w=x;
        m_h=y;
        width=x;
	height=y;
	this->xpos=xpos;
//	phx=x;
//	phy=y;

	printf("Creating SHM renderer for window %x, width %d, height %d\n", w, x, y);
	printf("Server vendor %s, release %d\n", ServerVendor(dpy), VendorRelease(dpy));
	
	try
	{
    	    XGCValues xcg;
	    xcg.graphics_exposures=false;
	    gc=XCreateGC(dpy, dev->handle(), GCGraphicsExposures, &xcg);
	    if ( !XShmQueryVersion(dpy, &major, &minor, &pixmaps_ok) )
		throw FATAL("MIT SHM extension not supported");

	    bool ok;
    	    xshminfo.shmid = shmget( IPC_PRIVATE,
    			     width*height*4/zoom/zoom,
    			     IPC_CREAT | 0777 );
	    ok = xshminfo.shmid != -1;
	    if(!ok)
		throw FATAL("Can't get shared memory segment");
    	
	    xshminfo.shmaddr = (char*)shmat( xshminfo.shmid, 0, 0 );
	    ok = xshminfo.shmaddr != 0;
	    xshminfo.readOnly = FALSE;

    	    if ( !ok )
		throw FATAL("Can't attach shared memory segment");
	    
	    ok = XShmAttach( dpy, &xshminfo );
	    if ( !ok )
		throw FATAL("XShmAttach failed");
	    
    	    xshmimg = XShmCreateImage( dpy, vis, dd, ZPixmap, xshminfo.shmaddr, &xshminfo, width/zoom, height/zoom );

	    if ( !xshmimg )
	        throw FATAL("Can't create shared image");

    	    printf("Video device: %d bpp ( physical %d ), color masks %X,%X,%X\n",
		dev->x11Depth(), GetPhysicalDepth(dev), vis->red_mask, vis->green_mask, vis->blue_mask);
//	pthread_mutex_init(&mutex, NULL);
    	    return 0;
	}
	catch(...)
	{
	    xshmimg = 0;
	    if ( xshminfo.shmaddr )
		shmdt( xshminfo.shmaddr );
	    if ( xshminfo.shmid != -1 )
		shmctl( xshminfo.shmid, IPC_RMID, 0 );
	    throw;
	}
    }
    virtual int Draw(QPainter* pm, const char* data)
    {
        char* outpic=xshmimg->data;
	if(outpic==0)
	    return 0;
	if(data==0)
	    return 0;

        Display *dpy = dev->x11Display();
    	Visual *vis = (Visual*)dev->x11Visual();

	int bit_depth=GetPhysicalDepth(dev);
	sync();
	const char* src;
	char* pic;
//	if(phx==::width && phy==::height)
	    pic=outpic;
//	else
//	    pic=new unsigned char[phx*phy*GetPhysicalDepth(dev)/8];
	
	if(zoom_divider==1)
		src=data;
	else
	{
		if(dev->x11Depth()==16)
		    zoom_2_to565((short*)outpic, (short*)data, ::width, ::height);
		else
		{
		    src=new char[::width*::height*4];
		    zoom_2((short*)src, (short*)data, ::width, ::height, bit_depth);
		}    
	}
	if(dev->x11Depth()==16)
	{
	    if(zoom_divider==1)
    		v555to565((short*)pic, (short*)src, ::width, ::height);
	}	
	else
	{	
		const char* tmp=src;
		bit_depth++;
    		bit_depth/=8;
		int bpl=::width*bit_depth/zoom_divider;
	    	pic+=bpl*(height/zoom_divider-1);

		for(int y=::height/zoom_divider-1; y>0; y--)
    		{
			memcpy(pic, tmp, bpl);
			tmp+=bpl;
			pic-=bpl;
    		}
        }
	if(zoom_divider!=1)
	    if(dev->x11Depth()!=16)
		delete src;
/*    
    	if(!(phx==::width && phy==::height))
	{
	    pthread_mutex_lock(&mutex);
	    AS_stretchBlit(pic, outpic);
	    pthread_mutex_unlock(&mutex);
	    delete pic;
	}		
*/			
	XShmPutImage(dpy, dev->handle(), gc,
		xshmimg, 0, 0, xpos, 0, ::width/zoom_divider, ::height/zoom_divider, true);
	return 0;
    }
/*    int Resize(int phw, int phh)
    {
	printf("ShmRenderer::Resize(%d,%d) from %d, %d\n", phw, phh, phx, phy);
	if(phw < ::width)return -1;
	if(phh < ::height)return -1;
	pthread_mutex_lock(&mutex);
	int result=AS_setupStretcher(::width,::height,phw,phh,GetPhysicalDepth(dev));
	phx=phw;
	phy=phh;
	pthread_mutex_unlock(&mutex);
	return result;
    }
*/
    virtual int sync()
    {
	Display *dpy = dev->x11Display();
        XSync(dpy, false);
	return 0;
    }
    virtual ~ShmRenderer()
    {
        if ( xshmimg == 0 )
	    return;
	Display *dpy = dev->x11Display();
        if ( xshmpm ) {
	    XFreePixmap( dpy, xshmpm );
	    xshmpm = 0;
        }
        XShmDetach( dpy, &xshminfo ); xshmimg->data = 0;
        qSafeXDestroyImage( xshmimg ); xshmimg = 0;
	shmdt( xshminfo.shmaddr );
        shmctl( xshminfo.shmid, IPC_RMID, 0 );
//	pthread_mutex_destroy(&mutex);
    }
};
#ifdef HAVE_LIBXXF86DGA
/********************************************************

	DGA video renderer
	Copyright 2000 Heinrich Langos

*********************************************************/
/*
class DGARenderer: public VideoRenderer
{
  int ws;
  char* base;
  int vwidth,vheight;
  char *conversionbuf;
public:
    DGARenderer(){}
    virtual int Create(QWidget* w, int x, int y, int zoom=1, int xpos=0)
    {
        dev=(QPaintDevice*)w;
	m_w=x;
        m_h=y;
        width=x;
	height=y;
        Display *dpy = dev->x11Display();
        int      bit_depth  = dev->x11Depth();
        Visual  *vis = (Visual*)dev->x11Visual();

	bool have_dga=false;
	conversionbuf=(char *)malloc(x*y*((bit_depth+7)/8));

	int minor, major, bank, ram, flags;
	if(XF86DGAQueryExtension(dpy, &major, &minor))
	{
    	    XF86DGAQueryDirectVideo(dpy, XDefaultScreen(dpy), &flags);
	    if(flags & XF86DGADirectPresent)
	    {
		have_dga=true;
		XF86DGAGetVideo(dpy, DefaultScreen(dpy),  &base, &ws, &bank, &ram);
		if(!base)
		    return -1;
	    }
	}

	if(have_dga) {

      XF86DGADirectVideo(dpy, DefaultScreen(dpy),
			 XF86DGADirectGraphics);
//			     XF86DGADirectMouse|
//			     XF86DGADirectKeyb);
	  //	  XF86DGAGetViewPortSize(dpy, DefaultScreen(dpy),
	  //				 &vwidth, &vheight);
	  //	  XF86DGASetViewPort(dpy, DefaultScreen(dpy), 0, 0);
	  return 0;
	}
    }
    virtual int Draw(QPainter* pm, const char* data)
    {
      Display *dpy = dev->x11Display();
      int bit_depth	 = dev->x11Depth();
      int byteppx = (bit_depth+7)/8;
	//	printf("memcpy-ing %d lines to base %x, dd: %d\n",height,base,dd);
	//	printf("ws: %d width: %d \n", ws, width);


	if((bit_depth==15)||(bit_depth==16)) {
//	    Visual* vis=(Visual*)dev->x11Visual();
//	    if((vis->red_mask==0xf800) && (vis->green_mask==0x7e0) && (vis->blue_mask==0x1F))
    	    v555to565((short*)conversionbuf, (short*)data, ::width, ::height);
//	    else
//	    {
//		printf("Unknown video mode: bpp %d, red %X, green %X, blue %X\n",
//		bit_depth, vis->red_mask, vis->green_mask, vis->blue_mask);
//		exit(1);
//	    }
	}
	else
        {
	    bit_depth++;
    	    bit_depth/=8;
	    int bpl=::width*bit_depth;
	    conversionbuf+=bpl*(height-1);
	    for(int y=::height-1; y>0; y--)
    	    {
		memcpy(conversionbuf, data, bpl);
		data+=bpl;
		conversionbuf-=bpl;
	    }
        }

	for(int i=0; i<height; i++)
	  memcpy((void*)(base+i*ws*byteppx), conversionbuf+i*width*byteppx, width*byteppx);
//	for(int i=0; i<height; i++)
//	  memcpy((void*)(base+i*ws*byteppx), data+i*width*bytepp, width*byteppx);
	return 0;
    };
    virtual int sync()
    {
	return 0;
    };
    virtual ~DGARenderer(){
        Display *dpy = dev->x11Display();
	XF86DGADirectVideo(dpy, DefaultScreen(dpy), 0);
	free(conversionbuf);
	}
};
*/
#endif  /*HAVE_LIBXXF86DGA*/

VideoRenderer* CreateRenderer()
{
    return new ShmRenderer;
}    
