/*
    Assembly image stretcher generator
    Copyright (C) 1999 Tomasz Cejner

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/    

#include "stretch.h"

#define CPU_RETN     0xC3
#define CPU_LODS     0xAD
#define CPU_STOS     0xAB
#define CPU_MOVS     0xA5
#define PREFIX16     0x66

typedef unsigned char Uint8;

int AS_setupStretcher (int src_w, int src_h, int dest_w, int dest_h, int bpp)
{
    int factor;
    int j,pos;
    Uint8 *eip;

    // better check bitdepth to avoid crash
    if (bpp!=16 && bpp!=32) return -1;
    if (dest_w<src_w || dest_h<src_h) return -1;

    // setup internal variables
    if (bpp==16) AS_src_w = src_w*2;
    if (bpp==32) AS_src_w = src_w*4;
    AS_dest_h = dest_h;
    AS_h_factor = (src_h<<16)/dest_h;

    // create scanline stretcher code
    eip = (Uint8*)AS_stretcher_code;
    
    factor = ( src_w << 16 )/dest_w;
    pos = 0;

    if (bpp==16) *eip++ = PREFIX16;
    *eip++ = CPU_LODS;
    
    for (j=0; j<dest_w; j++)
    {
        if ( pos >= 0x10000L )
        {
	    if (bpp==16) *eip++ = PREFIX16;
            *eip++ = CPU_LODS;
            pos -= 0x10000L;
        }
	if (bpp==16) *eip++ = PREFIX16;
        *eip++ = CPU_STOS;
        pos += factor;
	
    }
    *eip = CPU_RETN;
    return 0;
}
