/********************************************************

	Benchmarking tool
	Copyright 2000 Eugene Smith (divx@euro.ru)
	Last modified: 19.07.2000

*********************************************************/


#include <qapp.h>
#include <stdio.h>
#include <malloc.h>
#include <signal.h>
#include <iostream.h>
#include <unistd.h>
#include <avifile.h>
#include <qpainter.h>
//extern int EF_PROTECT_BELOW;
#include <aviplay.h>
#include "renderer.h"

#define __MODULE__ "benchmark"
inline unsigned int localcount()
{
    int a;
    __asm__ __volatile__("rdtsc\n\t"
    :"=a"(a)
    :
    :"edx");
    return a;
}
QWidget *m;
//QApplication* qApp;
int main(int argc, char **argv)
{
//    qApp=new QApplication(argc,argv);
    QApplication a(argc,argv);

//    EF_PROTECT_BELOW=1;
    if(GetAvifileVersion()!=AVIFILE_VERSION)
    {
	cout<<"This binary was compiled for Avifile ver. "<<AVIFILE_VERSION<<", but the library is ver. "<<GetAvifileVersion()<<". Aborting."<<endl;
	return 0;
    }	

    int result;
    char* fn;
    m=new QWidget;

    if(argc==1)
    {
	cout<<"Usage: benchmark <file> [-dga] [-zoom]"<<endl
	    <<"  Performs a measurement of Avifile performance for given file."<<endl
	    <<endl
	    <<"  Options: "<<endl
#ifdef HAVE_LIBXXF86DGA
	    <<"  -dga   Use incomplete DGA video renderer instead of default SHM one."<<endl
	    <<"			 Requires root privileges."<<endl
#endif
	    <<"  -zoom  Zoom picture to 50%"<<endl
	    <<" Description of result: "<<endl
	    <<"   \'Decompression\': Time spent on reading one video frame from file and its decompression."<<endl
	    <<"   \'Drawing\'      : Time spent on 15-to-16-bit conversion ( if it's needed ) and X server request."<<endl
	    <<"                         For DGA renderer includes time spent on actual drawing."<<endl
	    <<"   \'Sync\'         : Time spent by X server on drawing the frame. For DGA renderer it's zero because "<<endl
	    <<"                         all drawing is performed by Avifile."<<endl;
	return -1;
    }
    else
	fn=argv[1];
    int use_dga=0;
    int use_zoom=0;
    if(argc>=3)
    for(int i=2; i<argc; i++)
    {
#ifdef HAVE_LIBXXF86DGA
        if(strcmp(argv[i], "-dga")==0)
	    use_dga=1;
#endif
        if(strcmp(argv[i], "-zoom")==0)
	    use_zoom=1;
    }
    if(use_dga)use_zoom=0;//not implemented
    VideoRenderer* renderer=0;
    double read_times=0, draw_times=0, sync_times=0;
    int frames=0;
    MyPainter* painter=new MyPainter(m);
    IAviReadFile *file=NULL;
    try
    {

	file=CreateIAviReadFile(fn);
        IAviReadStream *stream=file->GetStream(0, AviStream::Video);
	
        BITMAPINFOHEADER bh;
	stream->StartStreaming();
        stream->SetBitDepth(GetPhysicalDepth(m));
        stream->GetOutputFormat(&bh, sizeof bh);
        if(stream->GetEndPos()<112)
	    throw FATAL("Too short movie");
#ifdef HAVE_LIBXXF86DGA
	if(use_dga)
	    renderer=CreateDGARenderer();
    	else
#endif	
        renderer=CreateShmRenderer();
        m->show();
	a.setMainWidget(m);
        printf("Movie size: %dx%d\n", bh.biWidth, bh.biHeight);
	renderer->Create(m,bh.biWidth,bh.biHeight, (use_zoom?2:1));
        m->move(100,100);
	m->resize(bh.biWidth/(use_zoom+1),bh.biHeight/(use_zoom+1));

	while(!stream->Eof() && (frames<32))
	{
	    stream->ReadFrame();
	    renderer->Draw(painter, stream->GetFrame());
    	    renderer->sync();
	    frames++;
	}		
        frames=0;
	while(!stream->Eof() && (frames<80))
        {
	    int t1, t2, t3, t4;
	    t1=localcount();
	    stream->ReadFrame();
	    t2=localcount();
	    renderer->Draw(painter, stream->GetFrame());
	    t3=localcount();
	    renderer->sync();
	    t4=localcount();
	    read_times+=(t2-t1)/550000.;
	    draw_times+=(t3-t2)/550000.;
	    sync_times+=(t4-t3)/550000.;
	    frames++;
	}		
        printf("Played %d frames in %f seconds ( max frame rate %f fps )\n",
	     frames, (read_times+draw_times+sync_times)/1000., 1000.*frames/(read_times+draw_times+sync_times));
	if(frames)
	printf("Average results:\n\t"
	    " Decompression %f ms\n\t"
	    " Drawing %f ms\n\t"
	    " Sync %f ms\n",
	    read_times/frames, draw_times/frames, sync_times/frames);
        if(use_dga==0)
	{
	    printf("Average video output speed: %f Mb/s\n",
	        GetPhysicalDepth(m)/8*bh.biWidth/(use_zoom+1)*bh.biHeight/(use_zoom+1)*frames/sync_times/1000.);
	}	    
        painter->end();
    }
    catch(FatalError& error)
    {
	error.Print();
    }		
    if(renderer)delete renderer;
    if(painter)delete painter;
    if(file)delete file;
    if(m)delete m;	
    return 0;
}